#! /usr/bin/python

#                   PCRE2 UNICODE PROPERTY SUPPORT
#                   ------------------------------

# This script generates the pcre2_ucp.h file from Unicode data files. This
# header uses enumerations to give names to Unicode property types and script
# names.

# This script was created in December 2021 as part of the Unicode data
# generation refactoring.


# Import common data lists and functions

from GenerateCommon import \
  bidi_classes, \
  bool_properties, \
  bool_props_list_item_size, \
  break_properties, \
  category_names, \
  general_category_names, \
  script_list_item_size, \
  script_names, \
  open_output

# Open the output file (no return on failure). This call also writes standard
# header boilerplate.

f = open_output("pcre2_ucp.h")

# Output this file's heading text

f.write("""\
#ifndef PCRE2_UCP_H_IDEMPOTENT_GUARD
#define PCRE2_UCP_H_IDEMPOTENT_GUARD

/* This file contains definitions of the Unicode property values that are
returned by the UCD access macros and used throughout PCRE2.

IMPORTANT: The specific values of the first two enums (general and particular
character categories) are assumed by the table called catposstab in the file
pcre2_auto_possess.c. They are unlikely to change, but should be checked after
an update. */
\n""")

f.write("/* These are the general character categories. */\n\nenum {\n")
for i in general_category_names:
  f.write("  ucp_%s,\n" % i)
f.write("};\n\n")

f.write("/* These are the particular character categories. */\n\nenum {\n")
for i in range(0, len(category_names), 2):
  f.write("  ucp_%s,    /* %s */\n" % (category_names[i], category_names[i+1]))
f.write("};\n\n")

f.write("/* These are Boolean properties. */\n\nenum {\n")
for i in bool_properties:
  f.write("  ucp_%s,\n" % i)

f.write("  /* This must be last */\n")
f.write("  ucp_Bprop_Count\n};\n\n")

f.write("/* Size of entries in ucd_boolprop_sets[] */\n\n")
f.write("#define ucd_boolprop_sets_item_size %d\n\n" % bool_props_list_item_size)

f.write("/* These are the bidi class values. */\n\nenum {\n")
for i in range(0, len(bidi_classes), 2):
  sp = ' ' * (4 - len(bidi_classes[i]))
  f.write("  ucp_bidi%s,%s /* %s */\n" % (bidi_classes[i], sp, bidi_classes[i+1]))
f.write("};\n\n")

f.write("/* These are grapheme break properties. The Extended Pictographic "
  "property\ncomes from the emoji-data.txt file. */\n\nenum {\n")
for i in range(0, len(break_properties), 2):
  sp = ' ' * (21 - len(break_properties[i]))
  f.write("  ucp_gb%s,%s /* %s */\n" % (break_properties[i], sp, break_properties[i+1]))
f.write("};\n\n")

f.write("/* These are the script identifications. */\n\nenum {\n  /* Scripts which has characters in other scripts. */\n")
for i in script_names:
  if i == "Unknown":
    f.write("\n  /* Scripts which has no characters in other scripts. */\n")
  f.write("  ucp_%s,\n" % i)
f.write("\n")

f.write("  /* This must be last */\n")
f.write("  ucp_Script_Count\n};\n\n")

f.write("/* Size of entries in ucd_script_sets[] */\n\n")
f.write("#define ucd_script_sets_item_size %d\n\n" % script_list_item_size)

f.write("#endif  /* PCRE2_UCP_H_IDEMPOTENT_GUARD */\n\n")
f.write("/* End of pcre2_ucp.h */\n")

f.close()

# End



# -*- coding:utf-8 -*-

from libc.stdint cimport uint8_t, uint32_t, int32_t


cdef extern from "pcre2.h":
    cdef unsigned int PCRE2_MAJOR
    cdef unsigned int PCRE2_MINOR

    # The following option bits can be passed to pcre2_compile(),
    # pcre2_match(), or pcre2_dfa_match(). PCRE2_NO_UTF_CHECK affects only the
    # function to which it is passed. Put these bits at the most significant
    # end of the options word so others can be added next to them.
    cdef unsigned int PCRE2_ANCHORED
    cdef unsigned int PCRE2_NO_UTF_CHECK
    cdef unsigned int PCRE2_ENDANCHORED

    # The following option bits can be passed only to pcre2_compile(). However,
    # they may affect compilation, JIT compilation, and/or interpretive
    # execution. The following tags indicate which:
    # C   alters what is compiled by pcre2_compile()
    # J   alters what is compiled by pcre2_jit_compile()
    # M   is inspected during pcre2_match() execution
    # D   is inspected during pcre2_dfa_match() execution
    cdef unsigned int PCRE2_ALLOW_EMPTY_CLASS    # C       
    cdef unsigned int PCRE2_ALT_BSUX             # C       
    cdef unsigned int PCRE2_AUTO_CALLOUT         # C       
    cdef unsigned int PCRE2_CASELESS             # C       
    cdef unsigned int PCRE2_DOLLAR_ENDONLY       #   J M D 
    cdef unsigned int PCRE2_DOTALL               # C       
    cdef unsigned int PCRE2_DUPNAMES             # C       
    cdef unsigned int PCRE2_EXTENDED             # C       
    cdef unsigned int PCRE2_FIRSTLINE            #   J M D 
    cdef unsigned int PCRE2_MATCH_UNSET_BACKREF  # C J M   
    cdef unsigned int PCRE2_MULTILINE            # C       
    cdef unsigned int PCRE2_NEVER_UCP            # C       
    cdef unsigned int PCRE2_NEVER_UTF            # C       
    cdef unsigned int PCRE2_NO_AUTO_CAPTURE      # C       
    cdef unsigned int PCRE2_NO_AUTO_POSSESS      # C       
    cdef unsigned int PCRE2_NO_DOTSTAR_ANCHOR    # C       
    cdef unsigned int PCRE2_NO_START_OPTIMIZE    #   J M D 
    cdef unsigned int PCRE2_UCP                  # C J M D 
    cdef unsigned int PCRE2_UNGREEDY             # C       
    cdef unsigned int PCRE2_UTF                  # C J M D 
    cdef unsigned int PCRE2_NEVER_BACKSLASH_C    # C       
    cdef unsigned int PCRE2_ALT_CIRCUMFLEX       #   J M D 
    cdef unsigned int PCRE2_ALT_VERBNAMES        # C       
    cdef unsigned int PCRE2_USE_OFFSET_LIMIT     #   J M D 
    cdef unsigned int PCRE2_EXTENDED_MORE        # C       
    cdef unsigned int PCRE2_LITERAL              # C       
    cdef unsigned int PCRE2_MATCH_INVALID_UTF    #   J M D

    # An additional compile options word is available in the compile context. 
    cdef unsigned int PCRE2_EXTRA_ALLOW_SURROGATE_ESCAPES  # C 
    cdef unsigned int PCRE2_EXTRA_BAD_ESCAPE_IS_LITERAL    # C 
    cdef unsigned int PCRE2_EXTRA_MATCH_WORD               # C 
    cdef unsigned int PCRE2_EXTRA_MATCH_LINE               # C 
    cdef unsigned int PCRE2_EXTRA_ESCAPED_CR_IS_LF         # C 
    cdef unsigned int PCRE2_EXTRA_ALT_BSUX                 # C 
    cdef unsigned int PCRE2_EXTRA_ALLOW_LOOKAROUND_BSK     # C 

    # These are for pcre2_jit_compile(). 
    cdef unsigned int PCRE2_JIT_COMPLETE  # For full matching.
    cdef unsigned int PCRE2_JIT_PARTIAL_SOFT
    cdef unsigned int PCRE2_JIT_PARTIAL_HARD
    cdef unsigned int PCRE2_JIT_INVALID_UTF

    # These are for pcre2_match(), pcre2_dfa_match(), pcre2_jit_match(), and
    # pcre2_substitute(). Some are allowed only for one of the functions, and
    # in these cases it is noted below. Note that PCRE2_ANCHORED,
    # PCRE2_ENDANCHORED and PCRE2_NO_UTF_CHECK can also be passed to these
    # functions (though pcre2_jit_match() ignores the latter since it bypasses
    # all sanity checks).
    cdef unsigned int PCRE2_NOTBOL
    cdef unsigned int PCRE2_NOTEOL
    cdef unsigned int PCRE2_NOTEMPTY          # ) These two must be kept
    cdef unsigned int PCRE2_NOTEMPTY_ATSTART  # ) adjacent to each other. 
    cdef unsigned int PCRE2_PARTIAL_SOFT
    cdef unsigned int PCRE2_PARTIAL_HARD
    cdef unsigned int PCRE2_DFA_RESTART  # pcre2_dfa_match() only 
    cdef unsigned int PCRE2_DFA_SHORTEST  # pcre2_dfa_match() only 
    cdef unsigned int PCRE2_SUBSTITUTE_GLOBAL  # pcre2_substitute() only 
    cdef unsigned int PCRE2_SUBSTITUTE_EXTENDED  # pcre2_substitute() only 
    cdef unsigned int PCRE2_SUBSTITUTE_UNSET_EMPTY  # pcre2_substitute() only 
    cdef unsigned int PCRE2_SUBSTITUTE_UNKNOWN_UNSET  # pcre2_substitute() only 
    cdef unsigned int PCRE2_SUBSTITUTE_OVERFLOW_LENGTH  # pcre2_substitute() only 
    cdef unsigned int PCRE2_NO_JIT  # Not for pcre2_dfa_match() 
    cdef unsigned int PCRE2_COPY_MATCHED_SUBJECT
    cdef unsigned int PCRE2_SUBSTITUTE_LITERAL  # pcre2_substitute() only 
    cdef unsigned int PCRE2_SUBSTITUTE_MATCHED  # pcre2_substitute() only 
    cdef unsigned int PCRE2_SUBSTITUTE_REPLACEMENT_ONLY  # pcre2_substitute() only 

    # Options for pcre2_pattern_convert(). 
    cdef unsigned int PCRE2_CONVERT_UTF
    cdef unsigned int PCRE2_CONVERT_NO_UTF_CHECK
    cdef unsigned int PCRE2_CONVERT_POSIX_BASIC
    cdef unsigned int PCRE2_CONVERT_POSIX_EXTENDED
    cdef unsigned int PCRE2_CONVERT_GLOB
    cdef unsigned int PCRE2_CONVERT_GLOB_NO_WILD_SEPARATOR
    cdef unsigned int PCRE2_CONVERT_GLOB_NO_STARSTAR

    # Newline and \R settings, for use in compile contexts. The newline values
    # must be kept in step with values set in config.h and both sets must all
    # be greater than zero.
    cdef int PCRE2_NEWLINE_CR
    cdef int PCRE2_NEWLINE_LF
    cdef int PCRE2_NEWLINE_CRLF
    cdef int PCRE2_NEWLINE_ANY
    cdef int PCRE2_NEWLINE_ANYCRLF
    cdef int PCRE2_NEWLINE_NUL

    cdef int PCRE2_BSR_UNICODE
    cdef int PCRE2_BSR_ANYCRLF

    # Error codes for pcre2_compile(). Some of these are also used by
    # pcre2_pattern_convert().
    cdef int PCRE2_ERROR_END_BACKSLASH
    cdef int PCRE2_ERROR_END_BACKSLASH_C
    cdef int PCRE2_ERROR_UNKNOWN_ESCAPE
    cdef int PCRE2_ERROR_QUANTIFIER_OUT_OF_ORDER
    cdef int PCRE2_ERROR_QUANTIFIER_TOO_BIG
    cdef int PCRE2_ERROR_MISSING_SQUARE_BRACKET
    cdef int PCRE2_ERROR_ESCAPE_INVALID_IN_CLASS
    cdef int PCRE2_ERROR_CLASS_RANGE_ORDER
    cdef int PCRE2_ERROR_QUANTIFIER_INVALID
    cdef int PCRE2_ERROR_INTERNAL_UNEXPECTED_REPEAT
    cdef int PCRE2_ERROR_INVALID_AFTER_PARENS_QUERY
    cdef int PCRE2_ERROR_POSIX_CLASS_NOT_IN_CLASS
    cdef int PCRE2_ERROR_POSIX_NO_SUPPORT_COLLATING
    cdef int PCRE2_ERROR_MISSING_CLOSING_PARENTHESIS
    cdef int PCRE2_ERROR_BAD_SUBPATTERN_REFERENCE
    cdef int PCRE2_ERROR_NULL_PATTERN
    cdef int PCRE2_ERROR_BAD_OPTIONS
    cdef int PCRE2_ERROR_MISSING_COMMENT_CLOSING
    cdef int PCRE2_ERROR_PARENTHESES_NEST_TOO_DEEP
    cdef int PCRE2_ERROR_PATTERN_TOO_LARGE
    cdef int PCRE2_ERROR_HEAP_FAILED
    cdef int PCRE2_ERROR_UNMATCHED_CLOSING_PARENTHESIS
    cdef int PCRE2_ERROR_INTERNAL_CODE_OVERFLOW
    cdef int PCRE2_ERROR_MISSING_CONDITION_CLOSING
    cdef int PCRE2_ERROR_LOOKBEHIND_NOT_FIXED_LENGTH
    cdef int PCRE2_ERROR_ZERO_RELATIVE_REFERENCE
    cdef int PCRE2_ERROR_TOO_MANY_CONDITION_BRANCHES
    cdef int PCRE2_ERROR_CONDITION_ASSERTION_EXPECTED
    cdef int PCRE2_ERROR_BAD_RELATIVE_REFERENCE
    cdef int PCRE2_ERROR_UNKNOWN_POSIX_CLASS
    cdef int PCRE2_ERROR_INTERNAL_STUDY_ERROR
    cdef int PCRE2_ERROR_UNICODE_NOT_SUPPORTED
    cdef int PCRE2_ERROR_PARENTHESES_STACK_CHECK
    cdef int PCRE2_ERROR_CODE_POINT_TOO_BIG
    cdef int PCRE2_ERROR_LOOKBEHIND_TOO_COMPLICATED
    cdef int PCRE2_ERROR_LOOKBEHIND_INVALID_BACKSLASH_C
    cdef int PCRE2_ERROR_UNSUPPORTED_ESCAPE_SEQUENCE
    cdef int PCRE2_ERROR_CALLOUT_NUMBER_TOO_BIG
    cdef int PCRE2_ERROR_MISSING_CALLOUT_CLOSING
    cdef int PCRE2_ERROR_ESCAPE_INVALID_IN_VERB
    cdef int PCRE2_ERROR_UNRECOGNIZED_AFTER_QUERY_P
    cdef int PCRE2_ERROR_MISSING_NAME_TERMINATOR
    cdef int PCRE2_ERROR_DUPLICATE_SUBPATTERN_NAME
    cdef int PCRE2_ERROR_INVALID_SUBPATTERN_NAME
    cdef int PCRE2_ERROR_UNICODE_PROPERTIES_UNAVAILABLE
    cdef int PCRE2_ERROR_MALFORMED_UNICODE_PROPERTY
    cdef int PCRE2_ERROR_UNKNOWN_UNICODE_PROPERTY
    cdef int PCRE2_ERROR_SUBPATTERN_NAME_TOO_LONG
    cdef int PCRE2_ERROR_TOO_MANY_NAMED_SUBPATTERNS
    cdef int PCRE2_ERROR_CLASS_INVALID_RANGE
    cdef int PCRE2_ERROR_OCTAL_BYTE_TOO_BIG
    cdef int PCRE2_ERROR_INTERNAL_OVERRAN_WORKSPACE
    cdef int PCRE2_ERROR_INTERNAL_MISSING_SUBPATTERN
    cdef int PCRE2_ERROR_DEFINE_TOO_MANY_BRANCHES
    cdef int PCRE2_ERROR_BACKSLASH_O_MISSING_BRACE
    cdef int PCRE2_ERROR_INTERNAL_UNKNOWN_NEWLINE
    cdef int PCRE2_ERROR_BACKSLASH_G_SYNTAX
    cdef int PCRE2_ERROR_PARENS_QUERY_R_MISSING_CLOSING
    # Error 159 is obsolete and should now never occur 
    cdef int PCRE2_ERROR_VERB_ARGUMENT_NOT_ALLOWED
    cdef int PCRE2_ERROR_VERB_UNKNOWN
    cdef int PCRE2_ERROR_SUBPATTERN_NUMBER_TOO_BIG
    cdef int PCRE2_ERROR_SUBPATTERN_NAME_EXPECTED
    cdef int PCRE2_ERROR_INTERNAL_PARSED_OVERFLOW
    cdef int PCRE2_ERROR_INVALID_OCTAL
    cdef int PCRE2_ERROR_SUBPATTERN_NAMES_MISMATCH
    cdef int PCRE2_ERROR_MARK_MISSING_ARGUMENT
    cdef int PCRE2_ERROR_INVALID_HEXADECIMAL
    cdef int PCRE2_ERROR_BACKSLASH_C_SYNTAX
    cdef int PCRE2_ERROR_BACKSLASH_K_SYNTAX
    cdef int PCRE2_ERROR_INTERNAL_BAD_CODE_LOOKBEHINDS
    cdef int PCRE2_ERROR_BACKSLASH_N_IN_CLASS
    cdef int PCRE2_ERROR_CALLOUT_STRING_TOO_LONG
    cdef int PCRE2_ERROR_UNICODE_DISALLOWED_CODE_POINT
    cdef int PCRE2_ERROR_UTF_IS_DISABLED
    cdef int PCRE2_ERROR_UCP_IS_DISABLED
    cdef int PCRE2_ERROR_VERB_NAME_TOO_LONG
    cdef int PCRE2_ERROR_BACKSLASH_U_CODE_POINT_TOO_BIG
    cdef int PCRE2_ERROR_MISSING_OCTAL_OR_HEX_DIGITS
    cdef int PCRE2_ERROR_VERSION_CONDITION_SYNTAX
    cdef int PCRE2_ERROR_INTERNAL_BAD_CODE_AUTO_POSSESS
    cdef int PCRE2_ERROR_CALLOUT_NO_STRING_DELIMITER
    cdef int PCRE2_ERROR_CALLOUT_BAD_STRING_DELIMITER
    cdef int PCRE2_ERROR_BACKSLASH_C_CALLER_DISABLED
    cdef int PCRE2_ERROR_QUERY_BARJX_NEST_TOO_DEEP
    cdef int PCRE2_ERROR_BACKSLASH_C_LIBRARY_DISABLED
    cdef int PCRE2_ERROR_PATTERN_TOO_COMPLICATED
    cdef int PCRE2_ERROR_LOOKBEHIND_TOO_LONG
    cdef int PCRE2_ERROR_PATTERN_STRING_TOO_LONG
    cdef int PCRE2_ERROR_INTERNAL_BAD_CODE
    cdef int PCRE2_ERROR_INTERNAL_BAD_CODE_IN_SKIP
    cdef int PCRE2_ERROR_NO_SURROGATES_IN_UTF16
    cdef int PCRE2_ERROR_BAD_LITERAL_OPTIONS
    cdef int PCRE2_ERROR_SUPPORTED_ONLY_IN_UNICODE
    cdef int PCRE2_ERROR_INVALID_HYPHEN_IN_OPTIONS
    cdef int PCRE2_ERROR_ALPHA_ASSERTION_UNKNOWN
    cdef int PCRE2_ERROR_SCRIPT_RUN_NOT_AVAILABLE
    cdef int PCRE2_ERROR_TOO_MANY_CAPTURES
    cdef int PCRE2_ERROR_CONDITION_ATOMIC_ASSERTION_EXPECTED
    cdef int PCRE2_ERROR_BACKSLASH_K_IN_LOOKAROUND

    # "Expected" matching error codes: no match and partial match. 
    cdef int PCRE2_ERROR_NOMATCH
    cdef int PCRE2_ERROR_PARTIAL

    # Error codes for UTF-8 validity checks.
    cdef int PCRE2_ERROR_UTF8_ERR1
    cdef int PCRE2_ERROR_UTF8_ERR2
    cdef int PCRE2_ERROR_UTF8_ERR3
    cdef int PCRE2_ERROR_UTF8_ERR4
    cdef int PCRE2_ERROR_UTF8_ERR5
    cdef int PCRE2_ERROR_UTF8_ERR6
    cdef int PCRE2_ERROR_UTF8_ERR7
    cdef int PCRE2_ERROR_UTF8_ERR8
    cdef int PCRE2_ERROR_UTF8_ERR9
    cdef int PCRE2_ERROR_UTF8_ERR10
    cdef int PCRE2_ERROR_UTF8_ERR11
    cdef int PCRE2_ERROR_UTF8_ERR12
    cdef int PCRE2_ERROR_UTF8_ERR13
    cdef int PCRE2_ERROR_UTF8_ERR14
    cdef int PCRE2_ERROR_UTF8_ERR15
    cdef int PCRE2_ERROR_UTF8_ERR16
    cdef int PCRE2_ERROR_UTF8_ERR17
    cdef int PCRE2_ERROR_UTF8_ERR18
    cdef int PCRE2_ERROR_UTF8_ERR19
    cdef int PCRE2_ERROR_UTF8_ERR20
    cdef int PCRE2_ERROR_UTF8_ERR21

    # Error codes for UTF-16 validity checks. 
    cdef int PCRE2_ERROR_UTF16_ERR1
    cdef int PCRE2_ERROR_UTF16_ERR2
    cdef int PCRE2_ERROR_UTF16_ERR3

    # Error codes for UTF-32 validity checks.
    cdef int PCRE2_ERROR_UTF32_ERR1
    cdef int PCRE2_ERROR_UTF32_ERR2

    # Miscellaneous error codes for pcre2[_dfa]_match(), substring extraction
    # functions, context functions, and serializing functions. They are in
    # numerical order. Originally they were in alphabetical order too, but now
    # that PCRE2 is released, the numbers must not be changed.
    cdef int PCRE2_ERROR_BADDATA
    cdef int PCRE2_ERROR_MIXEDTABLES  # Name was changed.
    cdef int PCRE2_ERROR_BADMAGIC
    cdef int PCRE2_ERROR_BADMODE
    cdef int PCRE2_ERROR_BADOFFSET
    cdef int PCRE2_ERROR_BADOPTION
    cdef int PCRE2_ERROR_BADREPLACEMENT
    cdef int PCRE2_ERROR_BADUTFOFFSET
    cdef int PCRE2_ERROR_CALLOUT  # Never used by PCRE2 itself.
    cdef int PCRE2_ERROR_DFA_BADRESTART
    cdef int PCRE2_ERROR_DFA_RECURSE
    cdef int PCRE2_ERROR_DFA_UCOND
    cdef int PCRE2_ERROR_DFA_UFUNC
    cdef int PCRE2_ERROR_DFA_UITEM
    cdef int PCRE2_ERROR_DFA_WSSIZE
    cdef int PCRE2_ERROR_INTERNAL
    cdef int PCRE2_ERROR_JIT_BADOPTION
    cdef int PCRE2_ERROR_JIT_STACKLIMIT
    cdef int PCRE2_ERROR_MATCHLIMIT
    cdef int PCRE2_ERROR_NOMEMORY
    cdef int PCRE2_ERROR_NOSUBSTRING
    cdef int PCRE2_ERROR_NOUNIQUESUBSTRING
    cdef int PCRE2_ERROR_NULL
    cdef int PCRE2_ERROR_RECURSELOOP
    cdef int PCRE2_ERROR_DEPTHLIMIT
    cdef int PCRE2_ERROR_RECURSIONLIMIT  # Obsolete synonym. 
    cdef int PCRE2_ERROR_UNAVAILABLE
    cdef int PCRE2_ERROR_UNSET
    cdef int PCRE2_ERROR_BADOFFSETLIMIT
    cdef int PCRE2_ERROR_BADREPESCAPE
    cdef int PCRE2_ERROR_REPMISSINGBRACE
    cdef int PCRE2_ERROR_BADSUBSTITUTION
    cdef int PCRE2_ERROR_BADSUBSPATTERN
    cdef int PCRE2_ERROR_TOOMANYREPLACE
    cdef int PCRE2_ERROR_BADSERIALIZEDDATA
    cdef int PCRE2_ERROR_HEAPLIMIT
    cdef int PCRE2_ERROR_CONVERT_SYNTAX
    cdef int PCRE2_ERROR_INTERNAL_DUPMATCH
    cdef int PCRE2_ERROR_DFA_UINVALID_UTF

    # Request types for pcre2_pattern_info().
    cdef int PCRE2_INFO_ALLOPTIONS
    cdef int PCRE2_INFO_ARGOPTIONS
    cdef int PCRE2_INFO_BACKREFMAX
    cdef int PCRE2_INFO_BSR
    cdef int PCRE2_INFO_CAPTURECOUNT
    cdef int PCRE2_INFO_FIRSTCODEUNIT
    cdef int PCRE2_INFO_FIRSTCODETYPE
    cdef int PCRE2_INFO_FIRSTBITMAP
    cdef int PCRE2_INFO_HASCRORLF
    cdef int PCRE2_INFO_JCHANGED
    cdef int PCRE2_INFO_JITSIZE
    cdef int PCRE2_INFO_LASTCODEUNIT
    cdef int PCRE2_INFO_LASTCODETYPE
    cdef int PCRE2_INFO_MATCHEMPTY
    cdef int PCRE2_INFO_MATCHLIMIT
    cdef int PCRE2_INFO_MAXLOOKBEHIND
    cdef int PCRE2_INFO_MINLENGTH
    cdef int PCRE2_INFO_NAMECOUNT
    cdef int PCRE2_INFO_NAMEENTRYSIZE
    cdef int PCRE2_INFO_NAMETABLE
    cdef int PCRE2_INFO_NEWLINE
    cdef int PCRE2_INFO_DEPTHLIMIT
    cdef int PCRE2_INFO_RECURSIONLIMIT  # Obsolete synonym 
    cdef int PCRE2_INFO_SIZE
    cdef int PCRE2_INFO_HASBACKSLASHC
    cdef int PCRE2_INFO_FRAMESIZE
    cdef int PCRE2_INFO_HEAPLIMIT
    cdef int PCRE2_INFO_EXTRAOPTIONS

    # Request types for pcre2_config(). 
    cdef int PCRE2_CONFIG_BSR
    cdef int PCRE2_CONFIG_JIT
    cdef int PCRE2_CONFIG_JITTARGET
    cdef int PCRE2_CONFIG_LINKSIZE
    cdef int PCRE2_CONFIG_MATCHLIMIT
    cdef int PCRE2_CONFIG_NEWLINE
    cdef int PCRE2_CONFIG_PARENSLIMIT
    cdef int PCRE2_CONFIG_DEPTHLIMIT
    cdef int PCRE2_CONFIG_RECURSIONLIMIT  # Obsolete synonym 
    cdef int PCRE2_CONFIG_STACKRECURSE  # Obsolete 
    cdef int PCRE2_CONFIG_UNICODE
    cdef int PCRE2_CONFIG_UNICODE_VERSION
    cdef int PCRE2_CONFIG_VERSION
    cdef int PCRE2_CONFIG_HEAPLIMIT
    cdef int PCRE2_CONFIG_NEVER_BACKSLASH_C
    cdef int PCRE2_CONFIG_COMPILED_WIDTHS
    cdef int PCRE2_CONFIG_TABLES_LENGTH


    # Opaque handles for PCRE2 defined structs.
    ctypedef struct pcre2_code_t "pcre2_code":
        pass
    ctypedef struct pcre2_match_data_t "pcre2_match_data":
        pass
    ctypedef struct pcre2_general_context_t "pcre2_general_context":
        pass
    ctypedef struct pcre2_compile_context_t "pcre2_compile_context":
        pass
    ctypedef struct pcre2_match_context_t "pcre2_match_context":
        pass

    # Basic string definition. Note that this assumes PCRE2 in compiled to
    # support 8-bit strings.
    ctypedef const uint8_t *pcre2_sptr_t "PCRE2_SPTR"

    
    # Error handling functions.
    int pcre2_get_error_message(
        int errorcode,
        uint8_t *buffer,
        size_t bufflen
    )

    # Pattern compilation functions.
    pcre2_code_t * pcre2_compile(
        pcre2_sptr_t pattern, 
        size_t length,
        uint32_t options,
        int *errorcode,
        size_t *erroroffset,
        pcre2_compile_context_t *ccontext
    )

    int pcre2_jit_compile(
        pcre2_code_t *code,
        uint32_t options
    )


    void pcre2_code_free(pcre2_code_t *code)

    # Information on compiled pattern.
    int pcre2_pattern_info(
        const pcre2_code_t *code,
        uint32_t what,
        void *where
    )

    int pcre2_substring_number_from_name(
        const pcre2_code_t *code,
        pcre2_sptr_t name
    )
    
    # Matching and match data functions.
    pcre2_match_data_t * pcre2_match_data_create(
        uint32_t ovecsize,
        pcre2_general_context_t *gcontext
    )
    
    pcre2_match_data_t * pcre2_match_data_create_from_pattern(
        const pcre2_code_t *code,
        pcre2_general_context_t *gcontext
    )
    
    int pcre2_match(
        const pcre2_code_t *code,
        pcre2_sptr_t subject,
        size_t length,
        size_t startoffset,
        uint32_t options,
        pcre2_match_data_t *match_data,
        pcre2_match_context_t *mcontext
    )
    int pcre2_jit_match(
        const pcre2_code_t *code,
        pcre2_sptr_t subject,
        size_t length,
        size_t startoffset,
        uint32_t options,
        pcre2_match_data_t *match_data,
        pcre2_match_context_t *mcontext
    )
    
    void pcre2_match_data_free(pcre2_match_data_t *match_data)

    uint32_t pcre2_get_ovector_count(pcre2_match_data_t *match_data)

    size_t *pcre2_get_ovector_pointer(pcre2_match_data_t *match_data)

    int pcre2_substring_nametable_scan(
        const pcre2_code_t *code,
        pcre2_sptr_t name,
        pcre2_sptr_t *first,
        pcre2_sptr_t *last
    )

    # String extraction from match data blocks.
    int pcre2_substring_get_byname(
        pcre2_match_data_t *match_data,
        pcre2_sptr_t name, 
        uint8_t **bufferptr,
        size_t *bufflen
    )

    int pcre2_substring_get_bynumber(
        pcre2_match_data_t *match_data,
        uint32_t number,
        uint8_t **bufferptr,
        size_t *bufflen
    )

    # Substitution.
    int pcre2_substitute(
        const pcre2_code_t *code,
        pcre2_sptr_t subject,
        size_t length,
        size_t startoffset,
        uint32_t options,
        pcre2_match_data_t *match_data,
        pcre2_match_context_t *mcontext,
        pcre2_sptr_t replacement,
        size_t rlength,
        uint8_t *outputbuffer,
        size_t *outlengthptr
    )

    # Serialization.
    int32_t pcre2_serialize_decode(
        pcre2_code_t **codes,
        int32_t number_of_codes,
        const uint8_t *code_bytes,
        pcre2_general_context_t *gcontex
    )
    int32_t pcre2_serialize_encode(
        pcre2_code_t **codes,
        int32_t number_of_codes,
        uint8_t **serialized_bytes,
        size_t *serialized_size,
        pcre2_general_context_t *gcontex
    )
    void pcre2_serialize_free(uint8_t *bytes)



# -*- coding:utf-8 -*-

# Standard libraries.
from libc.stdint cimport uint8_t


# Local imports.
from .utils cimport *
from .libpcre2 cimport *


class LibraryError(Exception):
    """ Catch all for other PCRE2 errors (e.g. bad option bits).
    """

    def __init__(self, errorcode, context_msg=""):
        cdef uint8_t errormsg_buf[120]
        get_error_message_rc = pcre2_get_error_message(
            errorcode, 
            errormsg_buf, sizeof(errormsg_buf)
        )

        # Handle errors in fetching error message.
        if get_error_message_rc == PCRE2_ERROR_NOMEMORY:
            raise MemoryError()
        elif get_error_message_rc < 0:
            raise LibraryError(
                get_error_message_rc,
                context_msg=f"Could not retrieve message for error code {get_error_message_rc}."
            )

        msg = errormsg_buf.decode("utf-8").capitalize()
        if context_msg:
            msg = msg + ". " + context_msg

        super().__init__(msg)
        self.errorcode = errorcode


class CompileError(LibraryError):
    """ Raised when pattern is malformed or is otherwise unable to be
    compiled.
    """
    
    def __init__(self, errorcode, context_msg=""):
        if not (errorcode > 0):
            raise ValueError("Compilation error codes are strictly positive")
        
        super().__init__(errorcode, context_msg=context_msg)


class MatchError(LibraryError):
    """ Raised when no or partial match found.
    """
    
    def __init__(self, errorcode, context_msg=""):
        if not (errorcode == PCRE2_ERROR_NOMATCH or errorcode == PCRE2_ERROR_PARTIAL):
            raise ValueError(
                f"Invalid error code '{errorcode}'. "
                "Match error codes can only be of value PCRE2_ERROR_NOMATCH or PCRE2_ERROR_PARTIAL"
            )
        
        super().__init__(errorcode, context_msg=context_msg)



#! /usr/bin/python

#                   PCRE2 UNICODE PROPERTY SUPPORT
#                   ------------------------------
#
# This script generates the pcre2_ucd.c file from Unicode data files. This is
# the compressed Unicode property data used by PCRE2. The script was created in
# December 2021 as part of the Unicode data generation refactoring. It is
# basically a re-working of the MultiStage2.py script that was submitted to the
# PCRE project by Peter Kankowski in 2008 as part of a previous upgrading of
# Unicode property support. A number of extensions have since been added. The
# main difference in the 2021 upgrade (apart from comments and layout) is that
# the data tables (e.g. list of script names) are now listed in or generated by
# a separate Python module that is shared with the other Generate scripts.
#
# This script must be run in the "maint" directory. It requires the following
# Unicode data tables: BidiMirrorring.txt, CaseFolding.txt,
# DerivedBidiClass.txt, DerivedCoreProperties.txt, DerivedGeneralCategory.txt,
# GraphemeBreakProperty.txt, PropList.txt, PropertyAliases.txt,
# PropertyValueAliases.txt, ScriptExtensions.txt, Scripts.txt, and
# emoji-data.txt. These must be in the Unicode.tables subdirectory.
#
# The emoji-data.txt file is found in the "emoji" subdirectory even though it
# is technically part of a different (but coordinated) standard as shown
# in files associated with Unicode Technical Standard #51 ("Unicode Emoji"),
# for example:
#
# http://unicode.org/Public/emoji/13.0/ReadMe.txt
#
# DerivedBidiClass.txt and DerivedGeneralCategory.txt are in the "extracted"
# subdirectory of the Unicode database (UCD) on the Unicode web site;
# GraphemeBreakProperty.txt is in the "auxiliary" subdirectory. The other files
# are in the top-level UCD directory.
#
# -----------------------------------------------------------------------------
# Minor modifications made to the original script:
#  Added #! line at start
#  Removed tabs
#  Made it work with Python 2.4 by rewriting two statements that needed 2.5
#  Consequent code tidy
#  Adjusted data file names to take from the Unicode.tables directory
#  Adjusted global table names by prefixing _pcre_.
#  Commented out stuff relating to the casefolding table, which isn't used;
#    removed completely in 2012.
#  Corrected size calculation
#  Add #ifndef SUPPORT_UCP to use dummy tables when no UCP support is needed.
#  Update for PCRE2: name changes, and SUPPORT_UCP is abolished.
#
# Major modifications made to the original script:
#  Added code to add a grapheme break property field to records.
#
#  Added code to search for sets of more than two characters that must match
#  each other caselessly. A new table is output containing these sets, and
#  offsets into the table are added to the main output records. This new
#  code scans CaseFolding.txt instead of UnicodeData.txt, which is no longer
#  used.
#
#  Update for Python3:
#    . Processed with 2to3, but that didn't fix everything
#    . Changed string.strip to str.strip
#    . Added encoding='utf-8' to the open() call
#    . Inserted 'int' before blocksize/ELEMS_PER_LINE because an int is
#        required and the result of the division is a float
#
#  Added code to scan the emoji-data.txt file to find the Extended Pictographic
#  property, which is used by PCRE2 as a grapheme breaking property. This was
#  done when updating to Unicode 11.0.0 (July 2018).
#
#  Added code to add a Script Extensions field to records. This has increased
#  their size from 8 to 12 bytes, only 10 of which are currently used.
#
#  Added code to add a bidi class field to records by scanning the
#  DerivedBidiClass.txt and PropList.txt files. This uses one of the two spare
#  bytes, so now 11 out of 12 are in use.
#
# 01-March-2010:     Updated list of scripts for Unicode 5.2.0
# 30-April-2011:     Updated list of scripts for Unicode 6.0.0
#     July-2012:     Updated list of scripts for Unicode 6.1.0
# 20-August-2012:    Added scan of GraphemeBreakProperty.txt and added a new
#                      field in the record to hold the value. Luckily, the
#                      structure had a hole in it, so the resulting table is
#                      not much bigger than before.
# 18-September-2012: Added code for multiple caseless sets. This uses the
#                      final hole in the structure.
# 30-September-2012: Added RegionalIndicator break property from Unicode 6.2.0
# 13-May-2014:       Updated for PCRE2
# 03-June-2014:      Updated for Python 3
# 20-June-2014:      Updated for Unicode 7.0.0
# 12-August-2014:    Updated to put Unicode version into the file
# 19-June-2015:      Updated for Unicode 8.0.0
# 02-July-2017:      Updated for Unicode 10.0.0
# 03-July-2018:      Updated for Unicode 11.0.0
# 07-July-2018:      Added code to scan emoji-data.txt for the Extended
#                      Pictographic property.
# 01-October-2018:   Added the 'Unknown' script name
# 03-October-2018:   Added new field for Script Extensions
# 27-July-2019:      Updated for Unicode 12.1.0
# 10-March-2020:     Updated for Unicode 13.0.0
# PCRE2-10.39:       Updated for Unicode 14.0.0
# 05-December-2021:  Added code to scan DerivedBidiClass.txt for bidi class,
#                      and also PropList.txt for the Bidi_Control property
# 19-December-2021:  Reworked script extensions lists to be bit maps instead
#                      of zero-terminated lists of script numbers.
# ----------------------------------------------------------------------------
#
# Changes to the refactored script:
#
# 26-December-2021:  Refactoring completed
# 10-January-2022:   Addition of general Boolean property support
# 12-January-2022:   Merge scriptx and bidiclass fields
# 14-January-2022:   Enlarge Boolean property offset to 12 bits
#
# ----------------------------------------------------------------------------
#
#
# The main tables generated by this script are used by macros defined in
# pcre2_internal.h. They look up Unicode character properties using short
# sequences of code that contains no branches, which makes for greater speed.
#
# Conceptually, there is a table of records (of type ucd_record), one for each
# Unicode character. Each record contains the script number, script extension
# value, character type, grapheme break type, offset to caseless matching set,
# offset to the character's other case, the bidi class, and offset to bitmap of
# Boolean properties.
#
# A real table covering all Unicode characters would be far too big. It can be
# efficiently compressed by observing that many characters have the same
# record, and many blocks of characters (taking 128 characters in a block) have
# the same set of records as other blocks. This leads to a 2-stage lookup
# process.
#
# This script constructs seven tables. The ucd_caseless_sets table contains
# lists of characters that all match each other caselessly. Each list is
# in order, and is terminated by NOTACHAR (0xffffffff), which is larger than
# any valid character. The first list is empty; this is used for characters
# that are not part of any list.
#
# The ucd_digit_sets table contains the code points of the '9' characters in
# each set of 10 decimal digits in Unicode. This is used to ensure that digits
# in script runs all come from the same set. The first element in the vector
# contains the number of subsequent elements, which are in ascending order.
#
# Scripts are partitioned into two groups. Scripts that appear in at least one
# character's script extension list come first, followed by "Unknown" and then
# all the rest. This sorting is done automatically in the GenerateCommon.py
# script. A script's number is its index in the script_names list.
#
# The ucd_script_sets table contains bitmaps that represent lists of scripts
# for Script Extensions properties. Each bitmap consists of a fixed number of
# unsigned 32-bit numbers, enough to allocate a bit for every script that is
# used in any character's extension list, that is, enough for every script
# whose number is less than ucp_Unknown. A character's script extension value
# in its ucd record is an offset into the ucd_script_sets vector. The first
# bitmap has no bits set; characters that have no script extensions have zero
# as their script extensions value so that they use this map.
#
# The ucd_boolprop_sets table contains bitmaps that represent lists of Boolean
# properties. Each bitmap consists of a fixed number of unsigned 32-bit
# numbers, enough to allocate a bit for each supported Boolean property.
#
# The ucd_records table contains one instance of every unique character record
# that is required. The ucd_stage1 table is indexed by a character's block
# number, which is the character's code point divided by 128, since 128 is the
# size of each block. The result of a lookup in ucd_stage1 a "virtual" block
# number.
#
# The ucd_stage2 table is a table of "virtual" blocks; each block is indexed by
# the offset of a character within its own block, and the result is the index
# number of the required record in the ucd_records vector.
#
# The following examples are correct for the Unicode 14.0.0 database. Future
# updates may make change the actual lookup values.
#
# Example: lowercase "a" (U+0061) is in block 0
#          lookup 0 in stage1 table yields 0
#          lookup 97 (0x61) in the first table in stage2 yields 35
#          record 35 is { 0, 5, 12, 0, -32, 18432, 44 }
#             0 = ucp_Latin   => Latin script
#             5 = ucp_Ll      => Lower case letter
#            12 = ucp_gbOther => Grapheme break property "Other"
#             0               => Not part of a caseless set
#           -32 (-0x20)       => Other case is U+0041
#         18432 = 0x4800      => Combined Bidi class + script extension values
#            44               => Offset to Boolean properties
#
# The top 5 bits of the sixth field are the Bidi class, with the rest being the
# script extension value, giving:
#
#             9 = ucp_bidiL   => Bidi class left-to-right
#             0               => No special script extension property
#
# Almost all lowercase latin characters resolve to the same record. One or two
# are different because they are part of a multi-character caseless set (for
# example, k, K and the Kelvin symbol are such a set).
#
# Example: hiragana letter A (U+3042) is in block 96 (0x60)
#          lookup 96 in stage1 table yields 93
#          lookup 66 (0x42) in table 93 in stage2 yields 819
#          record 819 is { 20, 7, 12, 0, 0, 18432, 82 }
#            20 = ucp_Hiragana => Hiragana script
#             7 = ucp_Lo       => Other letter
#            12 = ucp_gbOther  => Grapheme break property "Other"
#             0                => Not part of a caseless set
#             0                => No other case
#         18432 = 0x4800       => Combined Bidi class + script extension values
#            82                => Offset to Boolean properties
#
# The top 5 bits of the sixth field are the Bidi class, with the rest being the
# script extension value, giving:
#
#             9 = ucp_bidiL   => Bidi class left-to-right
#             0               => No special script extension property
#
# Example: vedic tone karshana (U+1CD0) is in block 57 (0x39)
#          lookup 57 in stage1 table yields 55
#          lookup 80 (0x50) in table 55 in stage2 yields 621
#          record 621 is { 84, 12, 3, 0, 0, 26762, 96 }
#            84 = ucp_Inherited => Script inherited from predecessor
#            12 = ucp_Mn        => Non-spacing mark
#             3 = ucp_gbExtend  => Grapheme break property "Extend"
#             0                 => Not part of a caseless set
#             0                 => No other case
#         26762 = 0x688A        => Combined Bidi class + script extension values
#            96                 => Offset to Boolean properties
#
# The top 5 bits of the sixth field are the Bidi class, with the rest being the
# script extension value, giving:
#
#            13 = ucp_bidiNSM   => Bidi class non-spacing mark
#           138                 => Script Extension list offset = 138
#
# At offset 138 in the ucd_script_sets vector we find a bitmap with bits 1, 8,
# 18, and 47 set. This means that this character is expected to be used with
# any of those scripts, which are Bengali, Devanagari, Kannada, and Grantha.
#
#  Philip Hazel, last updated 14 January 2022.
##############################################################################


# Import standard modules

import re
import string
import sys

# Import common data lists and functions

from GenerateCommon import \
  bidi_classes, \
  bool_properties, \
  bool_propsfiles, \
  bool_props_list_item_size, \
  break_properties, \
  category_names, \
  general_category_names, \
  script_abbrevs, \
  script_list_item_size, \
  script_names, \
  open_output

# Some general parameters

MAX_UNICODE = 0x110000
NOTACHAR = 0xffffffff


# ---------------------------------------------------------------------------
#                         DEFINE FUNCTIONS
# ---------------------------------------------------------------------------


# Parse a line of Scripts.txt, GraphemeBreakProperty.txt, DerivedBidiClass.txt
# or DerivedGeneralCategory.txt

def make_get_names(enum):
  return lambda chardata: enum.index(chardata[1])


# Parse a line of CaseFolding.txt

def get_other_case(chardata):
  if chardata[1] == 'C' or chardata[1] == 'S':
    return int(chardata[2], 16) - int(chardata[0], 16)
  return 0


# Parse a line of ScriptExtensions.txt

def get_script_extension(chardata):
  global last_script_extension

  offset = len(script_lists) * script_list_item_size
  if last_script_extension == chardata[1]:
    return offset - script_list_item_size

  last_script_extension = chardata[1]
  script_lists.append(tuple(script_abbrevs.index(abbrev) for abbrev in last_script_extension.split(' ')))
  return offset


# Read a whole table in memory, setting/checking the Unicode version

def read_table(file_name, get_value, default_value):
  global unicode_version

  f = re.match(r'^[^/]+/([^.]+)\.txt$', file_name)
  file_base = f.group(1)
  version_pat = r"^# " + re.escape(file_base) + r"-(\d+\.\d+\.\d+)\.txt$"
  file = open(file_name, 'r', encoding='utf-8')
  f = re.match(version_pat, file.readline())
  version = f.group(1)
  if unicode_version == "":
    unicode_version = version
  elif unicode_version != version:
    print("WARNING: Unicode version differs in %s", file_name, file=sys.stderr)

  table = [default_value] * MAX_UNICODE
  for line in file:
    line = re.sub(r'#.*', '', line)
    chardata = list(map(str.strip, line.split(';')))
    if len(chardata) <= 1:
      continue
    value = get_value(chardata)
    m = re.match(r'([0-9a-fA-F]+)(\.\.([0-9a-fA-F]+))?$', chardata[0])
    char = int(m.group(1), 16)
    if m.group(3) is None:
      last = char
    else:
      last = int(m.group(3), 16)
    for i in range(char, last + 1):
      # It is important not to overwrite a previously set value because in the
      # CaseFolding file there are lines to be ignored (returning the default
      # value of 0) which often come after a line which has already set data.
      if table[i] == default_value:
        table[i] = value
  file.close()
  return table


# Get the smallest possible C language type for the values in a table

def get_type_size(table):
  type_size = [("uint8_t", 1), ("uint16_t", 2), ("uint32_t", 4),
    ("signed char", 1), ("int16_t", 2), ("int32_t", 4)]
  limits = [(0, 255), (0, 65535), (0, 4294967295), (-128, 127),
    (-32768, 32767), (-2147483648, 2147483647)]
  minval = min(table)
  maxval = max(table)
  for num, (minlimit, maxlimit) in enumerate(limits):
    if minlimit <= minval and maxval <= maxlimit:
      return type_size[num]
  raise OverflowError("Too large to fit into C types")


# Get the total size of a list of tables

def get_tables_size(*tables):
  total_size = 0
  for table in tables:
    type, size = get_type_size(table)
    total_size += size * len(table)
  return total_size


# Compress a table into the two stages

def compress_table(table, block_size):
  blocks = {} # Dictionary for finding identical blocks
  stage1 = [] # Stage 1 table contains block numbers (indices into stage 2 table)
  stage2 = [] # Stage 2 table contains the blocks with property values
  table = tuple(table)
  for i in range(0, len(table), block_size):
    block = table[i:i+block_size]
    start = blocks.get(block)
    if start is None:
      # Allocate a new block
      start = len(stage2) / block_size
      stage2 += block
      blocks[block] = start
    stage1.append(start)
  return stage1, stage2


# Output a table

def write_table(table, table_name, block_size = None):
  type, size = get_type_size(table)
  ELEMS_PER_LINE = 16

  s = "const %s %s[] = { /* %d bytes" % (type, table_name, size * len(table))
  if block_size:
    s += ", block = %d" % block_size
  f.write(s + " */\n")
  table = tuple(table)
  if block_size is None:
    fmt = "%3d," * ELEMS_PER_LINE + " /* U+%04X */\n"
    mult = MAX_UNICODE / len(table)
    for i in range(0, len(table), ELEMS_PER_LINE):
      f.write(fmt % (table[i:i+ELEMS_PER_LINE] + (int(i * mult),)))
  else:
    if block_size > ELEMS_PER_LINE:
      el = ELEMS_PER_LINE
    else:
      el = block_size
    fmt = "%3d," * el + "\n"
    if block_size > ELEMS_PER_LINE:
      fmt = fmt * int(block_size / ELEMS_PER_LINE)
    for i in range(0, len(table), block_size):
      f.write(("\n/* block %d */\n" + fmt) % ((i / block_size,) + table[i:i+block_size]))
  f.write("};\n\n")


# Extract the unique combinations of properties into records

def combine_tables(*tables):
  records = {}
  index = []
  for t in zip(*tables):
    i = records.get(t)
    if i is None:
      i = records[t] = len(records)
    index.append(i)
  return index, records


# Create a record struct

def get_record_size_struct(records):
  size = 0
  structure = 'typedef struct {\n'
  for i in range(len(records[0])):
    record_slice = [record[i] for record in records]
    slice_type, slice_size = get_type_size(record_slice)
    # add padding: round up to the nearest power of slice_size
    size = (size + slice_size - 1) & -slice_size
    size += slice_size
    structure += '%s property_%d;\n' % (slice_type, i)

  # round up to the first item of the next structure in array
  record_slice = [record[0] for record in records]
  slice_type, slice_size = get_type_size(record_slice)
  size = (size + slice_size - 1) & -slice_size

  structure += '} ucd_record;\n*/\n'
  return size, structure


# Write records

def write_records(records, record_size):
  f.write('const ucd_record PRIV(ucd_records)[] = { ' + \
    '/* %d bytes, record size %d */\n' % (len(records) * record_size, record_size))
  records = list(zip(list(records.keys()), list(records.values())))
  records.sort(key = lambda x: x[1])
  for i, record in enumerate(records):
    f.write(('  {' + '%6d, ' * len(record[0]) + '}, /* %3d */\n') % (record[0] + (i,)))
  f.write('};\n\n')


# Write a bit set

def write_bitsets(list, item_size):
  for d in list:
    bitwords = [0] * item_size
    for idx in d:
      bitwords[idx // 32] |= 1 << (idx & 31)
    s = " "
    for x in bitwords:
      f.write("%s" % s)
      s = ", "
      f.write("0x%08xu" % x)
    f.write(",\n")
  f.write("};\n\n")


# ---------------------------------------------------------------------------
# This bit of code must have been useful when the original script was being
# developed. Retain it just in case it is ever needed again.

# def test_record_size():
#   tests = [ \
#     ( [(3,), (6,), (6,), (1,)], 1 ), \
#     ( [(300,), (600,), (600,), (100,)], 2 ), \
#     ( [(25, 3), (6, 6), (34, 6), (68, 1)], 2 ), \
#     ( [(300, 3), (6, 6), (340, 6), (690, 1)], 4 ), \
#     ( [(3, 300), (6, 6), (6, 340), (1, 690)], 4 ), \
#     ( [(300, 300), (6, 6), (6, 340), (1, 690)], 4 ), \
#     ( [(3, 100000), (6, 6), (6, 123456), (1, 690)], 8 ), \
#     ( [(100000, 300), (6, 6), (123456, 6), (1, 690)], 8 ), \
#   ]
#   for test in tests:
#     size, struct = get_record_size_struct(test[0])
#     assert(size == test[1])
# test_record_size()
# ---------------------------------------------------------------------------



# ---------------------------------------------------------------------------
#                       MAIN CODE FOR CREATING TABLES
# ---------------------------------------------------------------------------

unicode_version = ""

# Some of the tables imported from GenerateCommon.py have alternate comment
# strings for use by GenerateUcpHeader. The comments are not wanted here, so
# remove them.

bidi_classes = bidi_classes[::2]
break_properties = break_properties[::2]
category_names = category_names[::2]

# Create the various tables from Unicode data files

script = read_table('Unicode.tables/Scripts.txt', make_get_names(script_names), script_names.index('Unknown'))
category = read_table('Unicode.tables/DerivedGeneralCategory.txt', make_get_names(category_names), category_names.index('Cn'))
break_props = read_table('Unicode.tables/GraphemeBreakProperty.txt', make_get_names(break_properties), break_properties.index('Other'))
other_case = read_table('Unicode.tables/CaseFolding.txt', get_other_case, 0)
bidi_class = read_table('Unicode.tables/DerivedBidiClass.txt', make_get_names(bidi_classes), bidi_classes.index('L'))

# The grapheme breaking rules were changed for Unicode 11.0.0 (June 2018). Now
# we need to find the Extended_Pictographic property for emoji characters. This
# can be set as an additional grapheme break property, because the default for
# all the emojis is "other". We scan the emoji-data.txt file and modify the
# break-props table.

file = open('Unicode.tables/emoji-data.txt', 'r', encoding='utf-8')
for line in file:
  line = re.sub(r'#.*', '', line)
  chardata = list(map(str.strip, line.split(';')))
  if len(chardata) <= 1:
    continue
  if chardata[1] != "Extended_Pictographic":
    continue
  m = re.match(r'([0-9a-fA-F]+)(\.\.([0-9a-fA-F]+))?$', chardata[0])
  char = int(m.group(1), 16)
  if m.group(3) is None:
    last = char
  else:
    last = int(m.group(3), 16)
  for i in range(char, last + 1):
    if break_props[i] != break_properties.index('Other'):
      print("WARNING: Emoji 0x%x has break property %s, not 'Other'",
        i, break_properties[break_props[i]], file=sys.stderr)
    break_props[i] = break_properties.index('Extended_Pictographic')
file.close()

# Handle script extensions. The get_script_extesion() function maintains a
# list of unique bitmaps representing lists of scripts, returning the offset
# in that list. Initialize the list with an empty set, which is used for
# characters that have no script extensions.

script_lists = [[]]
last_script_extension = ""
scriptx_bidi_class = read_table('Unicode.tables/ScriptExtensions.txt', get_script_extension, 0)

for idx in range(len(scriptx_bidi_class)):
  scriptx_bidi_class[idx] = scriptx_bidi_class[idx] | (bidi_class[idx] << 11)
bidi_class = None

# Find the Boolean properties of each character. This next bit of magic creates
# a list of empty lists. Using [[]] * MAX_UNICODE gives a list of references to
# the *same* list, which is not what we want.

bprops = [[] for _ in range(MAX_UNICODE)]

# Collect the properties from the various files

for filename in bool_propsfiles:
  try:
    file = open('Unicode.tables/' + filename, 'r')
  except IOError:
    print(f"** Couldn't open {'Unicode.tables/' + filename}\n")
    sys.exit(1)

  for line in file:
    line = re.sub(r'#.*', '', line)
    data = list(map(str.strip, line.split(';')))
    if len(data) <= 1:
      continue

    try:
      ix = bool_properties.index(data[1])
    except ValueError:
      continue

    m = re.match(r'([0-9a-fA-F]+)(\.\.([0-9a-fA-F]+))?$', data[0])
    char = int(m.group(1), 16)
    if m.group(3) is None:
      last = char
    else:
      last = int(m.group(3), 16)

    for i in range(char, last + 1):
      bprops[i].append(ix)

  file.close()

# The ASCII property isn't listed in any files, but it is easy enough to add
# it manually.

ix = bool_properties.index("ASCII")
for i in range(128):
  bprops[i].append(ix)

# The Bidi_Mirrored property isn't listed in any property files. We have to
# deduce it from the file that lists the mirrored characters.

ix = bool_properties.index("Bidi_Mirrored")

try:
  file = open('Unicode.tables/BidiMirroring.txt', 'r')
except IOError:
  print(f"** Couldn't open {'Unicode.tables/BidiMirroring.txt'}\n")
  sys.exit(1)

for line in file:
  line = re.sub(r'#.*', '', line)
  data = list(map(str.strip, line.split(';')))
  if len(data) <= 1:
    continue
  c = int(data[0], 16)
  bprops[c].append(ix)

file.close()

# Scan each character's boolean property list and created a list of unique
# lists, at the same time, setting the index in that list for each property in
# the bool_props vector.

bool_props = [0] * MAX_UNICODE
bool_props_lists = [[]]

for c in range(MAX_UNICODE):
  s = set(bprops[c])
  for i in range(len(bool_props_lists)):
    if s == set(bool_props_lists[i]):
      break;
  else:
    bool_props_lists.append(bprops[c])
    i += 1

  bool_props[c] = i * bool_props_list_item_size

# This block of code was added by PH in September 2012. It scans the other_case
# table to find sets of more than two characters that must all match each other
# caselessly. Later in this script a table of these sets is written out.
# However, we have to do this work here in order to compute the offsets in the
# table that are inserted into the main table.

# The CaseFolding.txt file lists pairs, but the common logic for reading data
# sets only one value, so first we go through the table and set "return"
# offsets for those that are not already set.

for c in range(MAX_UNICODE):
  if other_case[c] != 0 and other_case[c + other_case[c]] == 0:
    other_case[c + other_case[c]] = -other_case[c]

# Now scan again and create equivalence sets.

caseless_sets = []

for c in range(MAX_UNICODE):
  o = c + other_case[c]

  # Trigger when this character's other case does not point back here. We
  # now have three characters that are case-equivalent.

  if other_case[o] != -other_case[c]:
    t = o + other_case[o]

    # Scan the existing sets to see if any of the three characters are already
    # part of a set. If so, unite the existing set with the new set.

    appended = 0
    for s in caseless_sets:
      found = 0
      for x in s:
        if x == c or x == o or x == t:
          found = 1

      # Add new characters to an existing set

      if found:
        found = 0
        for y in [c, o, t]:
          for x in s:
            if x == y:
              found = 1
          if not found:
            s.append(y)
        appended = 1

    # If we have not added to an existing set, create a new one.

    if not appended:
      caseless_sets.append([c, o, t])

# End of loop looking for caseless sets.

# Now scan the sets and set appropriate offsets for the characters.

caseless_offsets = [0] * MAX_UNICODE

offset = 1;
for s in caseless_sets:
  for x in s:
    caseless_offsets[x] = offset
  offset += len(s) + 1

# End of block of code for creating offsets for caseless matching sets.


# Combine all the tables

table, records = combine_tables(script, category, break_props,
  caseless_offsets, other_case, scriptx_bidi_class, bool_props)

# Find the record size and create a string definition of the structure for
# outputting as a comment.

record_size, record_struct = get_record_size_struct(list(records.keys()))

# Find the optimum block size for the two-stage table

min_size = sys.maxsize
for block_size in [2 ** i for i in range(5,10)]:
  size = len(records) * record_size
  stage1, stage2 = compress_table(table, block_size)
  size += get_tables_size(stage1, stage2)
  #print "/* block size %5d  => %5d bytes */" % (block_size, size)
  if size < min_size:
    min_size = size
    min_stage1, min_stage2 = stage1, stage2
    min_block_size = block_size


# ---------------------------------------------------------------------------
#                   MAIN CODE FOR WRITING THE OUTPUT FILE
# ---------------------------------------------------------------------------

# Open the output file (no return on failure). This call also writes standard
# header boilerplate.

f = open_output("pcre2_ucd.c")

# Output this file's heading text

f.write("""\
/* This file contains tables of Unicode properties that are extracted from
Unicode data files. See the comments at the start of maint/GenerateUcd.py for
details.

As well as being part of the PCRE2 library, this file is #included by the
pcre2test program, which redefines the PRIV macro to change table names from
_pcre2_xxx to xxxx, thereby avoiding name clashes with the library. At present,
just one of these tables is actually needed. When compiling the library, some
headers are needed. */

#ifndef PCRE2_PCRE2TEST
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "pcre2_internal.h"
#endif /* PCRE2_PCRE2TEST */

/* The tables herein are needed only when UCP support is built, and in PCRE2
that happens automatically with UTF support. This module should not be
referenced otherwise, so it should not matter whether it is compiled or not.
However a comment was received about space saving - maybe the guy linked all
the modules rather than using a library - so we include a condition to cut out
the tables when not needed. But don't leave a totally empty module because some
compilers barf at that. Instead, just supply some small dummy tables. */

#ifndef SUPPORT_UNICODE
const ucd_record PRIV(ucd_records)[] = {{0,0,0,0,0,0,0}};
const uint16_t PRIV(ucd_stage1)[] = {0};
const uint16_t PRIV(ucd_stage2)[] = {0};
const uint32_t PRIV(ucd_caseless_sets)[] = {0};
#else
\n""")

# --- Output some variable heading stuff ---

f.write("/* Total size: %d bytes, block size: %d. */\n\n" % (min_size, min_block_size))
f.write('const char *PRIV(unicode_version) = "{}";\n\n'.format(unicode_version))

f.write("""\
/* When recompiling tables with a new Unicode version, please check the types
in this structure definition with those in pcre2_internal.h (the actual field
names will be different).
\n""")

f.write(record_struct)

f.write("""
/* If the 32-bit library is run in non-32-bit mode, character values greater
than 0x10ffff may be encountered. For these we set up a special record. */

#if PCRE2_CODE_UNIT_WIDTH == 32
const ucd_record PRIV(dummy_ucd_record)[] = {{
  ucp_Unknown,    /* script */
  ucp_Cn,         /* type unassigned */
  ucp_gbOther,    /* grapheme break property */
  0,              /* case set */
  0,              /* other case */
  0 | (ucp_bidiL << UCD_BIDICLASS_SHIFT), /* script extension and bidi class */
  0,              /* bool properties offset */
  }};
#endif
\n""")

# --- Output the table of caseless character sets ---

f.write("""\
/* This table contains lists of characters that are caseless sets of
more than one character. Each list is terminated by NOTACHAR. */

const uint32_t PRIV(ucd_caseless_sets)[] = {
  NOTACHAR,
""")

for s in caseless_sets:
  s = sorted(s)
  for x in s:
    f.write('  0x%04x,' % x)
  f.write('  NOTACHAR,\n')
f.write('};\n\n')

# --- Other tables are not needed by pcre2test ---

f.write("""\
/* When #included in pcre2test, we don't need the table of digit sets, nor the
the large main UCD tables. */

#ifndef PCRE2_PCRE2TEST
\n""")

# --- Read Scripts.txt again for the sets of 10 digits. ---

digitsets = []
file = open('Unicode.tables/Scripts.txt', 'r', encoding='utf-8')

for line in file:
  m = re.match(r'([0-9a-fA-F]+)\.\.([0-9a-fA-F]+)\s+;\s+\S+\s+#\s+Nd\s+', line)
  if m is None:
    continue
  first = int(m.group(1),16)
  last  = int(m.group(2),16)
  if ((last - first + 1) % 10) != 0:
    f.write("ERROR: %04x..%04x does not contain a multiple of 10 characters" % (first, last),
      file=sys.stderr)
  while first < last:
    digitsets.append(first + 9)
    first += 10
file.close()
digitsets.sort()

f.write("""\
/* This table lists the code points for the '9' characters in each set of
decimal digits. It is used to ensure that all the digits in a script run come
from the same set. */

const uint32_t PRIV(ucd_digit_sets)[] = {
""")

f.write("  %d,  /* Number of subsequent values */" % len(digitsets))
count = 8
for d in digitsets:
  if count == 8:
    f.write("\n ")
    count = 0
  f.write(" 0x%05x," % d)
  count += 1
f.write("\n};\n\n")

f.write("""\
/* This vector is a list of script bitsets for the Script Extension property.
The number of 32-bit words in each bitset is #defined in pcre2_ucp.h as
ucd_script_sets_item_size. */

const uint32_t PRIV(ucd_script_sets)[] = {
""")
write_bitsets(script_lists, script_list_item_size)

f.write("""\
/* This vector is a list of bitsets for Boolean properties. The number of
32_bit words in each bitset is #defined as ucd_boolprop_sets_item_size in
pcre2_ucp.h. */

const uint32_t PRIV(ucd_boolprop_sets)[] = {
""")
write_bitsets(bool_props_lists, bool_props_list_item_size)


# Output the main UCD tables.

f.write("""\
/* These are the main two-stage UCD tables. The fields in each record are:
script (8 bits), character type (8 bits), grapheme break property (8 bits),
offset to multichar other cases or zero (8 bits), offset to other case or zero
(32 bits, signed), bidi class (5 bits) and script extension (11 bits) packed
into a 16-bit field, and offset in binary properties table (16 bits). */
\n""")

write_records(records, record_size)
write_table(min_stage1, 'PRIV(ucd_stage1)')
write_table(min_stage2, 'PRIV(ucd_stage2)', min_block_size)

f.write("#if UCD_BLOCK_SIZE != %d\n" % min_block_size)
f.write("""\
#error Please correct UCD_BLOCK_SIZE in pcre2_internal.h
#endif
#endif  /* SUPPORT_UNICODE */

#endif  /* PCRE2_PCRE2TEST */

/* End of pcre2_ucd.c */
""")

f.close

# End



# -*- coding:utf-8 -*-

# Standard libraries.
from libc.stdint cimport uint32_t
from libc.stdlib cimport malloc, free
from cpython cimport Py_buffer, PyBuffer_Release
from cpython cimport array
from cpython.unicode cimport PyUnicode_Check
from cpython.memoryview cimport PyMemoryView_FromMemory

# Local imports.
from .utils cimport *
from .libpcre2 cimport *
from .match cimport Match
from .scanner cimport Scanner
from .consts import BsrChar, NewlineChar


def _rebuild(pattern, code_bytes_obj, options):
    """ Deserializes code object to allow for unpickling.
    """
    patn = get_buffer(pattern)
    opts = <uint32_t>options
    code_buf = get_buffer(code_bytes_obj)
    
    cdef pcre2_code_t *code
    number_of_codes = pcre2_serialize_decode(&code, 1, <const uint8_t *>code_buf.buf, NULL)
    if number_of_codes < 0:
        raise_from_rc(number_of_codes, None)

    return Pattern._from_data(code, patn, opts)


cdef class Pattern:
    """
    Object wrapper for a compiled pattern (known as a code struct) in PCRE2.
    Attributes defined in pattern.pxd, see below for an overview:
        _code: Raw compiled pattern, managed by PCRE2
        _patn: Python object passed to compile
        _opts: Option bits passed to compile call
        _jitc: Indicator if pattern was JIT compiled
    """

    # =================================== #
    #         Lifetime management         #
    # =================================== #

    def __cinit__(self):
        self._code = NULL
        self._patn = NULL
        self._opts = 0
        self._jitc = False


    def __init__(self, *args, **kwargs):
        # Prevent accidental instantiation from normal Python code since we
        # cannot pass pointers into a Python constructor.
        module = self.__class__.__module__
        qualname = self.__class__.__qualname__
        raise TypeError(f"Cannot create '{module}.{qualname}' instances")


    def __dealloc__(self):
        if self._patn is not NULL:
            PyBuffer_Release(self._patn)
        if self._code is not NULL:
            pcre2_code_free(self._code)


    @staticmethod
    cdef Pattern _from_data(pcre2_code_t *code, Py_buffer *patn, uint32_t opts):
        """ Factory function to create Pattern objects from C-type fields. The
        ownership of the given pointers are stolen, which causes the extension
        type to free them when the object is deallocated.
        """
        # Fast call to __new__() that bypasses the __init__() constructor.
        cdef Pattern pattern = Pattern.__new__(Pattern)
        pattern._code = code
        pattern._patn = patn
        pattern._opts = opts
        return pattern


    # ========================================= #
    #         Serialize and deserialize         #
    # ========================================= #

    def __reduce__(self):
        """ Serializes code object to allow for pickling.
        """
        cdef uint8_t *code_bytes
        cdef size_t code_count
        serialize_rc = pcre2_serialize_encode(
            <const pcre2_code_t **>&self._code, 1, &code_bytes, &code_count, NULL
        )
        if serialize_rc < 0:
            raise_from_rc(serialize_rc, None)

        return (_rebuild, (self._patn.obj, code_bytes[:code_count], self._opts))


    # =================================== #
    #         Pattern information         #
    # =================================== #

    @staticmethod
    cdef uint32_t _info_uint(pcre2_code_t *code, uint32_t what) except *:
        """ Safely access pattern info returned as uint32_t.
        """
        cdef uint32_t where
        pattern_info_rc = pcre2_pattern_info(code, what, &where)
        if pattern_info_rc < 0:
            raise_from_rc(pattern_info_rc, None)
        return where

    @staticmethod
    cdef size_t _info_size(pcre2_code_t *code, uint32_t what) except *:
        """ Safely access pattern info returned as size_t.
        """
        cdef size_t where
        pattern_info_rc = pcre2_pattern_info(code, what, &where)
        if pattern_info_rc < 0:
            raise_from_rc(pattern_info_rc, None)
        return where

    @staticmethod
    cdef bint _info_bint(pcre2_code_t *code, uint32_t what) except *:
        """ Safely access pattern info returned as bint.
        """
        cdef bint where
        pattern_info_rc = pcre2_pattern_info(code, what, &where)
        if pattern_info_rc < 0:
            raise_from_rc(pattern_info_rc, None)
        return where


    @property
    def pattern(self):
        """ Return the pattern the object was compiled with.
        """
        return self._patn.obj


    @property
    def options(self):
        """ Returns the compile options as modified by any top-level (*XXX)
        option settings such as (*UTF) at the start of the pattern itself.
        """
        return Pattern._info_uint(self._code, PCRE2_INFO_ALLOPTIONS)


    @property
    def backslash_r(self):
        """ Return an indicator to what character sequences the \R escape
        sequence matches.
        """
        bsr = Pattern._info_uint(self._code, PCRE2_INFO_BSR)
        return BsrChar(bsr)


    @property
    def capture_count(self):
        """ Returns the highest capture group number in the pattern. In
        patterns where `(?|` is not used, this is also the total number of
        capture groups.
        """
        return Pattern._info_uint(self._code, PCRE2_INFO_CAPTURECOUNT)


    @property
    def jit_size(self):
        """ If the compiled pattern was successfully JIT compiled, return the
        size of the JIT compiled code, otherwise return zero.
        """
        return Pattern._info_size(self._code, PCRE2_INFO_JITSIZE)

    
    @property
    def name_count(self):
        """ Returns the number of named capture groups.
        """
        return Pattern._info_uint(self._code, PCRE2_INFO_NAMECOUNT)


    @property
    def newline(self):
        """ Returns the type of character sequence that will be recognized as
        a newline while matching.
        """
        newline = Pattern._info_uint(self._code, PCRE2_INFO_NEWLINE)
        return NewlineChar(newline)


    @property
    def size(self):
        """ Returns the size of the compiled pattern in bytes.
        """
        return Pattern._info_size(self._code, PCRE2_INFO_SIZE)


    def name_dict(self):
        """ Returns a mapping from capture group number to capture group name.
        """
        # Get name table related information.
        name_count = Pattern._info_uint(self._code, PCRE2_INFO_NAMECOUNT)
        name_entry_size = Pattern._info_uint(self._code, PCRE2_INFO_NAMEENTRYSIZE)

        cdef pcre2_sptr_t name_table
        pattern_info_rc = pcre2_pattern_info(self._code, PCRE2_INFO_NAMETABLE, &name_table)
        if pattern_info_rc < 0:
            raise_from_rc(pattern_info_rc, None)

        # Convert byte table to dictionary.
        name_dict = {}
        cdef uint32_t i
        for i in range(name_count):
            offset = i * name_entry_size

            # First two bytes of name table contain index, followed by possibly
            # unicode byte string.
            entry_idx = int((name_table[offset] << 8) | name_table[offset + 1])
            entry_name = name_table[offset + 2:offset + name_entry_size]

            # Clean up entry and convert to unicode as appropriate.
            entry_name = entry_name.strip(b"\x00")
            if PyUnicode_Check(self._patn.obj):
                entry_name = entry_name.decode("utf-8")

            name_dict[entry_idx] = entry_name

        return name_dict


    # ======================= #
    #         Methods         #
    # ======================= #

    def jit_compile(self):
        """ JIT compile the pattern.
        """
        jit_compile_rc = pcre2_jit_compile(self._code, PCRE2_JIT_COMPLETE)
        if jit_compile_rc < 0:
            raise_from_rc(jit_compile_rc, None)
        self._jitc = True

    
    @staticmethod
    cdef pcre2_match_data_t * _match(
            pcre2_code_t *code,
            Py_buffer *subj,
            size_t ofst,
            uint32_t opts,
            int *rc):
        """ Safe wrapper around calling PCRE2 function directly.
        """
        # Allocate memory for match.
        mtch = pcre2_match_data_create_from_pattern(code, NULL)
        if mtch is NULL:
            rc[0] = PCRE2_ERROR_NOMEMORY
            return NULL
        
        # Attempt match of pattern onto subject.
        rc[0] = pcre2_match(
            code, <pcre2_sptr_t>subj.buf, <size_t>subj.len,
            ofst, opts, mtch, NULL
        )
        return mtch


    def match(self, subject, offset=0, options=0):
        """ If match exists, returns the corresponding Match object. Otherwise
        a MatchError is thrown in the case of no matches. See the following
        PCRE2 documentation for a brief overview of the relevant options:
            http://pcre.org/current/doc/html/pcre2_match.html
        """
        cdef bint is_patn_utf = PyUnicode_Check(self._patn.obj)
        cdef bint is_subj_utf = PyUnicode_Check(subject)
        if is_patn_utf ^ is_subj_utf:
            patn_type = "string" if is_patn_utf else "bytes-like"
            subj_type = "string" if is_subj_utf else "bytes-like"
            raise ValueError(f"Cannot use a {patn_type} pattern with a {subj_type} subject")

        cdef Py_buffer *subj = get_buffer(subject)
        cdef size_t obj_ofst = <size_t>offset
        cdef size_t ofst = obj_ofst
        cdef uint32_t opts = <uint32_t>options

        # Convert indices accordingly.
        if is_subj_utf:
            ofst, obj_ofst = codepoint_to_codeunit(subj, obj_ofst, 0, 0)

        cdef int match_rc = 0
        mtch = Pattern._match(self._code, subj, ofst, opts, &match_rc)
        if match_rc < 0:
            raise_from_rc(match_rc, None)
            
        return Match._from_data(mtch, self, subj, ofst, opts)


    def scan(self, subject, offset=0):
        """ Returns iterator over all non-overlapping matches in a subject,
        yielding Match objects.
        """
        cdef bint is_patn_utf = PyUnicode_Check(self._patn.obj)
        cdef bint is_subj_utf = PyUnicode_Check(subject)
        if is_patn_utf ^ is_subj_utf:
            patn_type = "string" if is_patn_utf else "bytes-like"
            subj_type = "string" if is_subj_utf else "bytes-like"
            raise ValueError(f"Cannot use a {patn_type} pattern with a {subj_type} subject")

        return Scanner._from_data(self, subject, offset)


    @staticmethod
    cdef (uint8_t *, size_t) _substitute(
            pcre2_code_t *code,
            Py_buffer *repl,
            Py_buffer *subj,
            size_t res_buf_len,
            size_t ofst,
            uint32_t opts,
            pcre2_match_data_t *mtch,
            int *rc):
        """ Safe wrapper around calling PCRE2 function directly.
        """
        cdef size_t res_len = res_buf_len
        cdef uint8_t *res
        res = <uint8_t *>malloc(res_len * sizeof(uint8_t))
        substitute_rc = pcre2_substitute(
            code,
            <pcre2_sptr_t>subj.buf, <size_t>subj.len,
            ofst, opts | PCRE2_SUBSTITUTE_OVERFLOW_LENGTH, mtch, NULL,
            <pcre2_sptr_t>repl.buf, <size_t>repl.len,
            res, &res_len
        )
        # Reattempt substitution, now with required size of buffer known.
        if substitute_rc == PCRE2_ERROR_NOMEMORY:
            free(res)
            res = <uint8_t *>malloc(res_len * sizeof(uint8_t))
            substitute_rc = pcre2_substitute(
                code,
                <pcre2_sptr_t>subj.buf, <size_t>subj.len,
                ofst, opts, mtch, NULL,
                <pcre2_sptr_t>repl.buf, <size_t>repl.len,
                res, &res_len
            )
        # Capture return codes from both substitute attempts.
        if substitute_rc < 0:
            free(res)
            PyBuffer_Release(subj)
            PyBuffer_Release(repl)
            rc[0] = substitute_rc
            return NULL, 0
        
        return res, res_len


    def substitute(self, replacement, subject, offset=0, options=0, low_memory=False):
        """ Returns the string obtained by replaces matches in subject with a
        replacement. Note that option bits can significantly change the
        functions behavior. See the following PCRE2 documentation for a brief
        overview of the relevant options:
            http://pcre.org/current/doc/html/pcre2_substitute.html
        """
        is_patn_utf = <bint>PyUnicode_Check(self._patn.obj)
        is_subj_utf = <bint>PyUnicode_Check(subject)
        is_repl_utf = <bint>PyUnicode_Check(replacement)
        if is_subj_utf ^ is_repl_utf:
            subj_type = "string" if is_subj_utf else "bytes-like"
            repl_type = "string" if is_repl_utf else "bytes-like"
            raise ValueError(f"Cannot use a {subj_type} subject with a {repl_type} replacement")
        if is_patn_utf ^ is_subj_utf:
            patn_type = "string" if is_patn_utf else "bytes-like"
            subj_type = "string" if is_subj_utf else "bytes-like"
            raise ValueError(f"Cannot use a {patn_type} pattern with a {subj_type} subject")

        # Convert Python objects to C types.
        subj = get_buffer(subject)
        repl = get_buffer(replacement)
        cdef size_t obj_ofst = <size_t>offset
        cdef size_t ofst = obj_ofst
        cdef uint32_t opts = <uint32_t>options
        if is_subj_utf:
            ofst, obj_ofst = codepoint_to_codeunit(subj, obj_ofst, 0, 0)

        cdef size_t res_buf_len = 0
        if not low_memory:
            res_buf_len = subj.len + (subj.len // 2)

        cdef int rc = 0
        res, res_len = Pattern._substitute(
            self._code, repl, subj, res_buf_len, ofst, opts, NULL, &rc
        )
        if res is NULL:
            raise_from_rc(rc, None)

        # Clean up result and convert to unicode as appropriate.
        result = (<pcre2_sptr_t>res)[:res_len]
        result = result.strip(b"\x00")
        if is_subj_utf:
            result = result.decode("utf-8")
        
        free(res)
        PyBuffer_Release(subj)
        PyBuffer_Release(repl)
        return result
